from annet.annlib.lib import huawei_collapse_vlandb as collapse_vlandb
from annet.annlib.lib import huawei_expand_vlandb as expand_vlandb
from annet.annlib.types import Op

from annet.rulebook import common
from annet.rulebook.common import DiffItem


# =====
def single(rule, key, diff, **_):
    yield from _process_vlandb(rule, key, diff, False, False, None)


def multi(rule, key, diff, **_):
    yield from _process_vlandb(rule, key, diff, True, False, 10)


def multi_all(rule, key, diff, **_):
    yield from _process_vlandb(rule, key, diff, True, True, 10)


def vlan_diff(old, new, diff_pre, _pops):
    batch_new = set()  # vlan batch ... vlan ids
    for row in new:
        prefix, vlans = _parse_vlancfg(row)
        if prefix == "vlan batch":
            batch_new.update(vlans)
    ret = []
    for item in common.default_diff(old, new, diff_pre, _pops):
        prefix, vlan_ids = _parse_vlancfg(item.row)
        # если влан был объявлен глобально и при этом остается в батче
        # команда undo vlan ... будет пытаться полностью выпилить его с устройства
        # и из батча тоже. при этом делать undo vlan ... ; vlan batch ... не выход
        # поскольку для удаления cli требует удалить все vlanif"ы и проч
        if prefix == "vlan" and item.op == Op.REMOVED and batch_new.intersection(vlan_ids):
            result_item = DiffItem(Op.AFFECTED, item.row, item.children, item.diff_pre)
        # если влан объявлен глобально и одновременно с этим в батче
        # и при этом в блоке глобального объявления нет никаких опций
        # не добавляем его он будет висеть зазря - таким образом мы сохраним
        # симметрию с предыдущей логикой оба инварианта будут выдавать пустой патч
        elif prefix == "vlan" and batch_new.intersection(vlan_ids) and not item.children:
            result_item = None
        # vlan batch и остальное мы не трогаем
        else:
            result_item = item
        if result_item:
            ret.append(result_item)
    return ret


# =====
def _process_vlandb(rule, key, diff, multi, multi_all, multi_chunk):  # pylint: disable=unused-argument,redefined-outer-name
    assert len(diff[Op.AFFECTED]) == 0, "WTF? Affected signle: %r" % (diff[Op.AFFECTED])
    if not multi:
        for op in (Op.ADDED, Op.REMOVED):
            assert 0 <= len(diff[op]) <= 1, "Too many actions: %r" % (diff)

    if diff[Op.REMOVED] and not diff[Op.ADDED]:  # Removed
        if multi and multi_all:
            yield (False, rule["reverse"].format(*key) + " all", None)
            return
        elif not multi and not multi_all:
            yield (False, rule["reverse"].format(*key), None)
            return

    (prefix_add, new) = _parse_vlancfg_actions(diff[Op.ADDED])
    (prefix_del, old) = _parse_vlancfg_actions(diff[Op.REMOVED])
    removed = old.difference(new)
    added = new.difference(old)

    if removed:
        collapsed = collapse_vlandb(removed)
        for chunk in (_chunked(collapsed, multi_chunk) if multi else [collapsed]):
            yield (False, "undo %s %s" % (prefix_del, " ".join(chunk)), None)

    if added:
        collapsed = collapse_vlandb(added)
        for chunk in (_chunked(collapsed, multi_chunk) if multi else [collapsed]):
            yield (True, "%s %s" % (prefix_add, " ".join(chunk)), None)


def _chunked(items, size):
    for offset in range(0, len(items), size):
        yield items[offset:offset + size]


def _parse_vlancfg_actions(actions):
    prefix = None
    vlandb = set()
    for action in actions:
        (prefix, part) = _parse_vlancfg(action["row"])
        vlandb.update(part)
    return (prefix, vlandb)


def _parse_vlancfg(row):
    parts = row.split()
    assert len(parts) > 0, row
    index = None
    for (index, item) in reversed(list(enumerate(parts))):
        if not (item.isdigit() or item == "to"):
            break
    prefix = " ".join(parts[:index + 1])
    vlandb = expand_vlandb(" ".join(parts[index + 1:]))
    return (prefix, vlandb)


def _find_new_vlans(root_pre):
    ret = set()
    for (rule, pre) in root_pre.items():
        if not rule.startswith("vlan batch"):
            continue
        new = _parse_vlancfg_actions(pre["items"][tuple()][Op.ADDED])[1]
        ret.update(new)
    return ret



from annet.annlib.tabparser import (  # pylint: disable=unused-import
    AsrFormatter,
    CiscoFormatter,
    CommonFormatter,
    HuaweiFormatter,
    JuniperFormatter,
    JuniperList,
    NokiaFormatter,
    ParserError,
    RibbonFormatter,
    RosFormatter,
    parse_to_tree,
)


def make_formatter(hw, **kwargs):
    if hw.Huawei:
        cls = HuaweiFormatter
    elif hw.Cisco.ASR or hw.Cisco.XRV:
        cls = AsrFormatter
    elif hw.Nexus or hw.Cisco or hw.Arista or hw.Aruba:
        cls = CiscoFormatter
    elif hw.Juniper:
        cls = JuniperFormatter
    elif hw.Nokia:
        cls = NokiaFormatter
    elif hw.RouterOS:
        cls = RosFormatter
    elif hw.PC:
        cls = CommonFormatter
    elif hw.Ribbon:
        cls = RibbonFormatter
    else:
        raise NotImplementedError("Unknown formatter for hw '%s'" % hw)

    return cls(**kwargs)



# pylint: disable=too-many-ancestors

import abc
import argparse
import enum
import logging
import os

from valkit.common import valid_string_list

from annet.argparse import Arg, ArgGroup, DefaultFromEnv
from annet.hardware import hardware_connector
from annet.storage import Query, storage_connector


# ====
def valid_vendor(vendor):
    hw_provider = hardware_connector.get()
    hw = hw_provider.vendor_to_hw(vendor)
    if hw:
        return hw.vendor
    return ""


def convert_to_none(arg):
    return None if arg == "-" else arg


def valid_config_source(value):
    if value not in ["cfglister", "running", "empty", "-"] and not os.path.exists(value):
        raise ValueError("No such file or directory %r" % value)
    return value


def valid_range(value: str):
    if value.isdigit():
        return slice(0, int(value))
    elif ":" in value:
        start_str, stop_str = value.split(":", 1)
        if not stop_str:
            stop = None
        else:
            stop = int(stop_str)
        return slice(int(start_str), stop)

    raise ValueError("Invalid range: %s" % value)


def opt_query_factory(**kwargs):
    return Arg(
        "query",
        help="Запрос, определяющий список устройств. Принимается fqdn, rackcode, глоб"
             " или путь к файлу со списком запросов (нужно писать @path/to/file)",
        **kwargs,
    )


# ====
opt_query = opt_query_factory(nargs="+")

opt_query_optional = opt_query_factory(nargs="*", default=[])

opt_dest = Arg(
    "--dest", type=convert_to_none,
    help="Файл или каталог для вывода сгенерированных данных"
)

opt_expand_path = Arg(
    "--expand-path", default=False,
    help="Разворачивать пути entire-генераторов при записи их на файловую систему"
)

opt_old = Arg(
    "old",
    help="Файл со старым конфигом (или каталог с пачкой)"
)

opt_new = Arg(
    "new",
    help="Файл с новым конфигом (или каталог с пачкой)"
)

opt_hw = Arg(
    "--hw", default="", type=valid_vendor,
    help="Производитель устройства (например Huawei или Cisco) или полное название модели. Если его нет - пытаемся его задетектить"
)

opt_indent = Arg(
    "--indent", default="  ",
    help="Отступ при форматировании блоков"
)

opt_allowed_gens = Arg(
    "-g", "--allowed-gens", type=valid_string_list,
    help="Список классов генераторов через запятую, которые нужно запустить"
)

opt_excluded_gens = Arg(
    "-G", "--excluded-gens", type=valid_string_list,
    help="Список классов генераторов через запятую, запуск которых следует исключить"
)

opt_force_enabled = Arg(
    "--force-enabled", type=valid_string_list,
    help="Список классов генераторов через запятую, которые не нужно считать DISABLED даже при наличии тега"
)

opt_generators_context = Arg(
    "--generators_context", type=str, default=None,
    help=argparse.SUPPRESS
)

opt_no_acl = Arg(
    "--no-acl", default=False,
    help="Отключение ACL при генерации"
)

opt_no_acl_exclusive = Arg(
    "--no-acl-exclusive", default=False,
    help="Проверяем что ACL выполненных генераторов не пересекаются"
)

opt_acl_safe = Arg(
    "--acl-safe", default=False,
    help="Использовать более строгий safe acl для фильтрации результата генерации"
)

opt_show_rules = Arg(
    "--show-rules", default=False,
    help="Показывать правила rulebook при выводе диффа"
)

opt_tolerate_fails = Arg(
    "--tolerate-fails", default=False,
    help="Рапортовать об ошибках без остановки генерации"
)

# При параллельном запуске и включённом --tolerate-fails код возврата
# всегда нулевой. Это не позволяет нам легко понять, прошла ли генерация
# успешно для всех устройств. С этим флажком код будет ненулевой, если
# генерация упала хотя бы для одного устройства. А в хелпе эту переменную
# не выводим, там и так не протолкнуться от флагов.
opt_strict_exit_code = Arg(
    "--strict-exit-code", default=False,
    help=argparse.SUPPRESS,
)

opt_required_packages_check = Arg(
    "--required-packages-check", default=False,
    help="Включить проверку наличия установленных deb-пакетов для Entire-генераторов"
)

opt_profile = Arg(
    "--profile", default=False,
    help="Показать в stderr время, затраченное на работу генераторов и обращениям к RackTables"
)


opt_parallel = Arg(
    "-P", "--parallel", type=int, default=1,
    help="Количество одновременных потоков генерирования"
)

opt_max_tasks = Arg(
    "--max-tasks", type=int, default=None,
    help="Рестартовать воркеры каждые N устройств, для сброса кеша и ограничения потребления памяти"
         "По умолчанию - не рестартовать"
)

opt_annotate = Arg(
    "--annotate", default=False,
    help="Добавить к сгенерированному конфигу комментарии о том откуда строчка взялась"
)

opt_config = Arg(
    "--config", default="running", type=valid_config_source,
    help="'cfglister', 'running', 'empty', путь к файлу конфига, "
         "каталогу с файлами конфига в формате <hostname>.cfg "
         "или '-' (stdin)"
)

opt_clear = Arg(
    "--clear", default=False,
    help="Используя acl вычищает команды относящиеся к данному генератору"
         "аналогично использованию return в самом начале генератора"
)

opt_filter_acl = Arg(
    "--filter-acl", default="",
    help="путь к файлу с дополнительным фильтрующим acl, или '-' (stdin)"
)

opt_filter_ifaces = Arg(
    "-i", "--filter-ifaces", default=[], type=valid_string_list,
    help="Генерирует filter-acl по имени интерфейса. "
         "Принимает регекспы, через запятую: '-i 10GE,100GE'. "
         "По-умолчанию добавляет '.*' к концу каждого. "
         "Для указания имени точно, следует добавлять '$': '-i ae0$'. "
         "Если filter-acl передан напрямую, данная опция игнорируется."
)

opt_filter_peers = Arg(
    "-fp", "--filter-peers", default=[], type=valid_string_list,
    help="Генерирует filter-acl по адресу/имени группы/дескрипшену пира."
)

opt_filter_policies = Arg(
    "-frp", "--filter-policies", default=[], type=valid_string_list,
    help="Генерирует filter-acl по названию политик, название должно строго соответствовать, частичные имена не пройдут"
)

opt_ask_pass = Arg(
    "--ask-pass", default=False,
    help="Спросить пароль на подключение"
)

opt_no_ask_deploy = Arg(
    "--no-ask-deploy", default=False,
    help="Не подтвеждать команды перед выполнением"
)

opt_no_progress = Arg(
    "--no-progress", default=False,
    help="Выключить графику прогресс баров комокутора"
)

opt_log_json = Arg(
    "--log-json", default=False,
    help="Логгировать в формате json (default: plain text)"
)

opt_log_dest = Arg(
    "--log-dest", default="deploy/",
    help="Логгировать в указанный файл/директорию, или в stdout, если указать '-'"
)

opt_log_nogroup = Arg(
    "--log-nogroup", default=False,
    help="Не создавать в директории LOG-DEST поддиректории DATE_TIME/"
)

opt_max_slots = Arg(
    "--max-slots", default=30, type=int,
    help="Количество одновременно обрабатываемых asyncio устройств"
)

opt_hosts_range = Arg(
    "--hosts-range", type=valid_range,
    help="Обработать только указанный диапазон хостов. 10 - первые 10. 10:20 - хосты с 10-го по 20-ый"
)

opt_add_comments = Arg(
    "--add-comments", default=False,
    help="Добавлять комменты подтверждения для rbprocess"
)

opt_no_label = Arg(
    "--no-label", default=False,
    help="Убрать лейбл с именем файла из вывода"
)

opt_no_color = Arg(
    "--no-color", default=False,
    help="Не делать ANSI-раскраску вывода (при --dest включён)"
)

opt_no_check_diff = Arg(
    "--no-check-diff", default=False,
    help="не запрашивать дифф после деплоя"
)

opt_dont_commit = Arg(
    "--dont-commit", default=False,
    help="не добавлять команду commit во время деплоя"
)

opt_rollback = Arg(
    "--rollback", default=False,
    help="предложить откат после деплоя где это возможно"
)

opt_fail_on_empty_config = Arg(
    "--fail-on-empty-config", default=False,
    help=argparse.SUPPRESS,
)


opt_show_generators_format = Arg(
    "--format", default="text", choices=["text", "json"],
    help="Формат выдачи"
)


class EntireReloadFlag(enum.Enum):
    no = "no"
    yes = "yes"
    force = "force"

    def __bool__(self):
        return self is not self.no

    def __str__(self):
        return str(self.value)

    __repr__ = __str__


opt_entire_reload = Arg(
    "--entire-reload",
    type=EntireReloadFlag,
    default=EntireReloadFlag.yes,
    choices=list(EntireReloadFlag),
    const=EntireReloadFlag.yes,
    nargs="?",
    help="Выполнить reload() при deploy'e entire генераторов. "
    "no/yes/force - нет/только если файл изменился/даже если не изменился"
)

opt_show_hosts_progress = Arg(
    "--show-hosts-progress", default=False,
    help="Показывать проценты выполнения по хостам"
)

opt_no_collapse = Arg(
    "--no-collapse", default=False,
    help="Не схлопывать одинаковые diff для группы устройств (при --dest включён)"
)

opt_fails_only = Arg(
    "--fails-only", default=False,
    help="Показать только устройства с ошибками"
)

opt_connect_timeout = Arg(
    "--connect-timeout", default=DefaultFromEnv("ANN_CONNECT_TIMEOUT", "20.0"), type=float,
    help="Таймаут на подключение к устройству в секундах."
         " Значение по-умолчанию можно задать в переменной окружения ANN_CONNECT_TIMEOUT"
)

opt_selected_context_name = Arg(
    "context-name", type=str, help="Имя контекста в файле конфигурации"
)


# ====
class CacheOptions(ArgGroup):
    no_mesh = False


class QueryOptionsBase(CacheOptions):
    @property
    @abc.abstractmethod
    def query(self) -> "Query":
        pass

    @property
    @abc.abstractmethod
    def hosts_range(self):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(self.query, Query):
            connectors = storage_connector.get_all()
            if not connectors:
                pass
            elif len(connectors) == 1:
                query_type = connectors[0].query()
                self.query = query_type.new(self.query, hosts_range=self.hosts_range)
            else:
                logging.warning("Multiple connectors found, skip parsing query")


class QueryOptions(QueryOptionsBase):
    query = opt_query
    hosts_range = opt_hosts_range


class QueryOptionsOptional(QueryOptionsBase):
    query = opt_query_optional
    hosts_range = opt_hosts_range


class ParallelOptions(ArgGroup):
    parallel = opt_parallel
    max_tasks = opt_max_tasks


class GenSelectOptions(ArgGroup):
    allowed_gens = opt_allowed_gens
    excluded_gens = opt_excluded_gens
    force_enabled = opt_force_enabled
    generators_context = opt_generators_context
    ignore_disabled = False


class GenOptions(QueryOptions, GenSelectOptions, CacheOptions, ParallelOptions):
    no_acl = opt_no_acl
    no_acl_exclusive = opt_no_acl_exclusive
    acl_safe = opt_acl_safe
    filter_acl = opt_filter_acl
    filter_ifaces = opt_filter_ifaces
    filter_peers = opt_filter_peers
    filter_policies = opt_filter_policies
    profile = opt_profile
    tolerate_fails = opt_tolerate_fails
    required_packages_check = opt_required_packages_check
    strict_exit_code = opt_strict_exit_code
    fail_on_empty_config = opt_fail_on_empty_config


class ComocutorOptions(ArgGroup):
    ask_pass = opt_ask_pass
    max_slots = opt_max_slots
    no_progress = opt_no_progress
    connect_timeout = opt_connect_timeout


class CliLoggingOptions(ArgGroup):
    log_json = opt_log_json
    log_dest = opt_log_dest
    log_nogroup = opt_log_nogroup


class DeviceCliOptions(ComocutorOptions, CliLoggingOptions):
    no_ask_deploy = opt_no_ask_deploy
    dont_commit = opt_dont_commit


class FileOutOptions(ArgGroup):
    dest = opt_dest
    expand_path = opt_expand_path
    no_label = opt_no_label
    no_color = opt_no_color

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.dest:
            self.no_color = True


class DiffOptions(GenOptions, ComocutorOptions):
    clear = opt_clear
    config = opt_config


class FileInputOptions(ArgGroup):
    old = opt_old
    new = opt_new
    hw = opt_hw
    fails_only = opt_fails_only


class PatchOptions(DiffOptions):
    add_comments = opt_add_comments


class ShowGenOptions(GenOptions, FileOutOptions):
    indent = opt_indent
    annotate = opt_annotate
    show_hosts_progress = opt_show_hosts_progress


class ShowDiffOptions(DiffOptions, FileOutOptions):
    indent = opt_indent
    show_rules = opt_show_rules
    no_collapse = opt_no_collapse

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.dest:
            self.no_collapse = True


class ShowPatchOptions(PatchOptions, FileOutOptions):
    indent = opt_indent
    show_hosts_progress = opt_show_hosts_progress


class FileDiffOptions(FileInputOptions, FileOutOptions, ParallelOptions):
    indent = opt_indent
    show_rules = opt_show_rules


class FilePatchOptions(FileInputOptions, FileOutOptions, ParallelOptions):
    indent = opt_indent
    add_comments = opt_add_comments


class ShowGeneratorsOptions(QueryOptionsOptional, GenSelectOptions):
    format = opt_show_generators_format
    acl_safe = opt_acl_safe
    ignore_disabled = False


class DeployOptions(ShowDiffOptions, PatchOptions, DeviceCliOptions):
    no_check_diff = opt_no_check_diff
    entire_reload = opt_entire_reload
    rollback = opt_rollback


class SelectContext(ArgGroup):
    context_name = opt_selected_context_name



import argparse
import operator
import os
import platform
import subprocess
import shutil
from contextlib import ExitStack, contextmanager
from typing import Tuple, Iterable

import yaml
from contextlog import get_logger
from valkit.python import valid_logging_level

from annet.deploy import driver_connector, fetcher_connector
from annet import api, cli_args, filtering
from annet.api import collapse_texts, Deployer
from annet.argparse import ArgParser, subcommand
from annet.diff import gen_sort_diff
from annet.gen import Loader, old_raw
from annet.lib import get_context_path, repair_context_file
from annet.output import output_driver_connector, OutputDriver
from annet.storage import storage_connector


def fill_base_args(parser: ArgParser, pkg_name: str, logging_config: str):
    parser.add_argument("--log-level", default="WARN", type=valid_logging_level,
                        help="Уровень детализации логов (DEBUG, DEBUG2 (with comocutor debug), INFO, WARN, CRITICAL)")
    parser.add_argument("--pkg_name", default=pkg_name, help=argparse.SUPPRESS)
    parser.add_argument("--logging_config", default=logging_config, help=argparse.SUPPRESS)


def list_subcommands():
    return globals().copy()


def _gen_current_items(
        config,
        stdin,
        loader: Loader,
        output_driver: OutputDriver,
        gen_args: cli_args.GenOptions,
) -> Iterable[Tuple[str, str, bool]]:
    for device, result in old_raw(
        args=gen_args,
        loader=loader,
        config=config,
        stdin=stdin,
        do_files_download=True,
        use_mesh=False,
    ):
        if device.hw.vendor != "pc":
            destname = output_driver.cfg_file_names(device)[0]
            yield (destname, result, False)
        else:
            for entire_path, entire_data in sorted(result.items(), key=operator.itemgetter(0)):
                if entire_data is None:
                    entire_data = ""
                destname = output_driver.entire_config_dest_path(device, entire_path)
                yield (destname, entire_data, False)


@contextmanager
def get_loader(gen_args: cli_args.GenOptions, args: cli_args.QueryOptions):
    exit_stack = ExitStack()
    connectors = storage_connector.get_all()
    storages = []
    with exit_stack:
        for connector in connectors:
            storage_opts = connector.opts().from_cli_opts(args)
            storages.append(exit_stack.enter_context(connector.storage()(storage_opts)))
        yield Loader(*storages, args=gen_args)


@subcommand(cli_args.QueryOptions, cli_args.opt_config, cli_args.FileOutOptions)
def show_current(args: cli_args.QueryOptions, config, arg_out: cli_args.FileOutOptions) -> None:
    """ Показать текущий конфиг устройств """
    gen_args = cli_args.GenOptions(args, no_acl=True)
    output_driver = output_driver_connector.get()
    with get_loader(gen_args, args) as loader:
        if not loader.devices:
            get_logger().error("No devices found for %s", args.query)

        items = _gen_current_items(
            loader=loader,
            output_driver=output_driver,
            gen_args=gen_args,
            stdin=args.stdin(config=config),
            config=config,
        )
        output_driver.write_output(arg_out, items, len(loader.devices))


@subcommand(cli_args.ShowGenOptions)
def gen(args: cli_args.ShowGenOptions):
    """ Сгенерировать конфиг для устройств """
    with get_loader(args, args) as loader:
        (success, fail) = api.gen(args, loader)

        out = [item for items in success.values() for item in items]
        output_driver = output_driver_connector.get()
        if args.dest is None:
            text_mapping = {item[0]: item[1] for item in out}
            out = [(",".join(key), value, False) for key, value in collapse_texts(text_mapping).items()]

        out.extend(output_driver.format_fails(fail, loader.device_fqdns))
        total = len(success) + len(fail)
        if not total:
            get_logger().error("No devices found for %s", args.query)
        output_driver.write_output(args, out, total)


@subcommand(cli_args.ShowDiffOptions)
def diff(args: cli_args.ShowDiffOptions):
    """ Сгенерировать конфиг для устройств и показать дифф по рулбуку с текущим """
    with get_loader(args, args) as loader:
        filterer = filtering.filterer_connector.get()
        device_ids = loader.device_ids
        output_driver_connector.get().write_output(
            args,
            gen_sort_diff(api.diff(args, loader, device_ids, filterer), args),
            len(loader.device_ids)
        )


@subcommand(cli_args.ShowPatchOptions)
def patch(args: cli_args.ShowPatchOptions):
    """ Сгенерировать конфиг для устройств и сформировать патч """
    with get_loader(args, args) as loader:
        (success, fail) = api.patch(args, loader)

        out = [item for items in success.values() for item in items]
        output_driver = output_driver_connector.get()
        out.extend(output_driver.format_fails(fail, loader.device_fqdns))
        total = len(success) + len(fail)
        if not total:
            get_logger().error("No devices found for %s", args.query)
        output_driver.write_output(args, out, total)


@subcommand(cli_args.DeployOptions)
def deploy(args: cli_args.DeployOptions):
    """ Сгенерировать конфиг для устройств и задеплоить его """

    deployer = Deployer(args)
    filterer = filtering.filterer_connector.get()
    fetcher = fetcher_connector.get()
    deploy_driver = driver_connector.get()

    with get_loader(args, args) as loader:
        return api.deploy(
            args=args, loader=loader, deployer=deployer,
            deploy_driver=deploy_driver, filterer=filterer,
            fetcher=fetcher,
        )


@subcommand(cli_args.FileDiffOptions)
def file_diff(args: cli_args.FileDiffOptions):
    """ Показать дифф по рулбуку между файлами или каталогами """
    (success, fail) = api.file_diff(args)
    out = []
    output_driver = output_driver_connector.get()
    if not args.fails_only:
        out.extend(item for items in success.values() for item in items)
    out.extend(output_driver.format_fails(fail))
    # todo отрефакторить логику с отображением хоста в диффе: передавать в write_output явно критерий
    output_driver.write_output(args, out, len(out) + 1)


@subcommand(cli_args.FilePatchOptions)
def file_patch(args: cli_args.FilePatchOptions):
    """ Сформировать патч для файлов или каталогов """
    (success, fail) = api.file_patch(args)
    out = []
    output_driver = output_driver_connector.get()
    if not args.fails_only:
        out.extend(item for items in success.values() for item in items)
    out.extend(output_driver.format_fails(fail))
    output_driver.write_output(args, out, len(out))


@subcommand()
def context():
    """ Операции для управления файлом контекста.

    По-умолчанию находится в '~/.annushka/context.yml', либо по пути в переменной окружения ANN_CONTEXT_CONFIG_PATH.
    """
    context_touch()


@subcommand(parent=context)
def context_touch():
    """ Вывести путь к файлу контекста и, при отсутствии, создать его и наполнить данными (команда по-умолчанию) """
    print(get_context_path(touch=True))


@subcommand(cli_args.SelectContext, parent=context)
def context_set_context(args: cli_args.SelectContext):
    """ Задать текущий активный контекст по имени в конфигурации

    Выбранный контекст будет использоваться по-умолчанию при незаданной переменной окружения ANN_SELECTED_CONTEXT
    """
    with open(path := get_context_path(touch=True)) as f:
        data = yaml.safe_load(f)
    if args.context_name not in data.get("context", {}):
        raise KeyError(f"Cannot select context with name '{args.context_name}'. "
                       f"Available options are: {list(data.get('context', []))}")
    data["selected_context"] = args.context_name
    with open(path, "w") as f:
        yaml.dump(data, f, sort_keys=False)


@subcommand(parent=context)
def context_edit():
    """ Открыть файл конфигурации контекста в редакторе из переменной окружения EDITOR

    Если переменная окружения EDITOR не задана,
    для Windows пытаемся открыть файл средствами ОС, для остальных случаев пытаемся открыть в vi
    """
    editor = ""
    if e := os.getenv("EDITOR"):
        editor = e
    elif platform.system() == "Windows":
        editor = "notepad.exe"
    elif shutil.which("vim"):
        editor = "vim"
    else:
        editor = "vi"
    path = get_context_path(touch=True)
    proc = subprocess.Popen([editor, path])
    proc.wait()


@subcommand(parent=context)
def context_repair():
    """ Попытаться исправить расхождения в формате файла контекста после изменении версии """
    repair_context_file()



import functools
import re
from collections import OrderedDict as odict

from valkit.common import valid_bool

from . import platform, syntax


# =====
@functools.lru_cache()
def compile_ordering_text(text, vendor):
    return _compile_ordering(
        tree=syntax.parse_text(text, params_scheme={
            "order_reverse": {
                "validator": valid_bool,
                "default":   False,
            },
        }),
        reverse_prefix=platform.VENDOR_REVERSES[vendor],
    )


def decompile_ordering_rulebook(rb) -> str:
    def _decompile_ordering_text(rb, level):
        indent = "  "
        for attrs in rb.values():
            yield indent * level + attrs["attrs"]["raw_rule"]
            yield from _decompile_ordering_text(attrs["children"], level + 1)
    return "\n".join(_decompile_ordering_text(rb, 0))


# =====
def _compile_ordering(tree, reverse_prefix):
    ordering = odict()
    for (rule_id, attrs) in tree.items():
        if attrs["type"] == "normal":
            ordering[rule_id] = {
                "attrs": {
                    "direct_regexp": syntax.compile_row_regexp(attrs["row"]),
                    "reverse_regexp": (
                        syntax.compile_row_regexp(reverse_prefix + " " + attrs["row"])
                        if not attrs["row"].startswith(reverse_prefix + " ") else
                        syntax.compile_row_regexp(re.sub(r"^%s\s+" % (reverse_prefix), "", attrs["row"]))
                    ),
                    "order_reverse": attrs["params"]["order_reverse"],
                    "raw_rule": attrs["raw_rule"],
                    "context": attrs["context"],
                },
                "children": _compile_ordering(attrs["children"], reverse_prefix),
            }
    return ordering



import sys
from abc import ABC
from functools import cached_property
from importlib.metadata import entry_points
from typing import Generic, Optional, Type, TypeVar, List

T = TypeVar("T")


class Connector(ABC, Generic[T]):
    name: str
    ep_name: str
    ep_group: str = "annet.connectors"
    _classes: Optional[List[Type[T]]] = None

    def _get_default(self) -> Type[T]:
        raise RuntimeError(f"{self.name} is not set")

    @cached_property
    def _entry_point(self) -> List[Type[T]]:
        return load_entry_point(self.ep_group, self.ep_name)

    def get(self, *args, **kwargs) -> T:
        if self._classes is None:
            self._classes = self._entry_point or [self._get_default()]
        if len(self._classes) > 1:
            raise RuntimeError(
                f"Multiple classes are registered with the same "
                f"group={self.ep_group} and name={self.ep_name}: "
                f"{[cls for cls in self._classes]}",
            )

        res = self._classes[0]
        return res(*args, **kwargs)

    def get_all(self, *args, **kwargs) -> T:
        if self._classes is None:
            self._classes = self._entry_point or [self._get_default()]

        return [cls(*args, **kwargs) for cls in self._classes]

    def set(self, cls: Type[T]):
        if self._classes is not None:
            raise RuntimeError(f"Cannot reinitialize value of {self.name}")
        self._classes = [cls]

    def set_all(self, classes: List[Type[T]]):
        if self._classes is not None:
            raise RuntimeError(f"Cannot reinitialize value of {self.name}")
        self._classes = list(classes)

    def is_default(self) -> bool:
        return self._classes is self._entry_point is None


class CachedConnector(Connector[T], ABC):
    _cache: Optional[T] = None

    def get(self, *args, **kwargs) -> T:
        assert not (args or kwargs), "Arguments forwarding is not allowed for cached connectors"
        if self._cache is None:
            self._cache = super().get()
        return self._cache

    def set(self, cls: Type[T]):
        super().set(cls)
        self._cache = None


def load_entry_point(group: str, name: str):
    if sys.version_info < (3, 10):
        ep = [item for item in entry_points().get(group, []) if item.name == name]
    else:
        ep = entry_points(group=group, name=name)  # pylint: disable=unexpected-keyword-arg
    if not ep:
        return None
    return [item.load() for item in ep]



import functools
import re
from collections import OrderedDict as odict
from collections import namedtuple

Answer = namedtuple("Answer", ("text", "send_nl"))


def compile_messages(tree):
    ignore = []
    dialogs = odict()
    for attrs in tree.values():
        if attrs["type"] == "normal":
            match = re.match(r"^ignore:(.+)$", attrs["row"])
            if match:
                ignore.append(MakeMessageMatcher(match.group(1)))
                continue

            match = re.match(r"^dialog:(.+):::(.+)$", attrs["row"])
            if match:
                dialogs[MakeMessageMatcher(match.group(1))] = Answer(
                    text=match.group(2).strip(),
                    send_nl=attrs["params"]["send_nl"],
                )
    return (ignore, dialogs)


class MakeMessageMatcher:
    def __init__(self, text):
        text = text.strip()
        self._text = text
        if text.startswith("/") and text.endswith("/"):
            regexp = re.compile(text[1:-1].strip(), flags=re.I)
            self._fn = regexp.match
        else:
            self._fn = (lambda arg: _simplify_text(text) in _simplify_text(arg))

    def __str__(self):
        return "%s(%r)" % (self.__class__.__name__, self._text)

    __repr__ = __str__

    def __call__(self, intext):
        return self._fn(intext)

    def __eq__(self, other):
        return type(other) is type(self) and self._text == other._text  # pylint: disable=protected-access

    def __hash__(self):
        return hash("%s_%s" % (self.__class__.__name__, self._text))


@functools.lru_cache()
def _simplify_text(text):
    return re.sub(r"\s", "", text).lower()



from logging import getLogger
from typing import Optional, List, Union, Dict
from ipaddress import ip_interface
from collections import defaultdict

from adaptix import P
from adaptix.conversion import impl_converter, link
from annetbox.v37 import models as api_models
from annetbox.v37.client_sync import NetboxV37

from annet.adapters.netbox.common import models
from annet.adapters.netbox.common.manufacturer import (
    is_supported, get_hw, get_breed,
)
from annet.adapters.netbox.common.query import NetboxQuery
from annet.adapters.netbox.common.storage_opts import NetboxStorageOpts
from annet.annlib.netdev.views.hardware import HardwareView
from annet.storage import Storage


logger = getLogger(__name__)


@impl_converter(recipe=[
    link(P[api_models.Device].name, P[models.NetboxDevice].hostname),
    link(P[api_models.Device].name, P[models.NetboxDevice].fqdn),
])
def extend_device_base(
        device: api_models.Device,
        interfaces: List[models.Interface],
        hw: Optional[HardwareView],
        breed: str,
        neighbours: Optional[List[models.NetboxDevice]],
        storage: Storage,
) -> models.NetboxDevice:
    ...


def extend_device(
        device: api_models.Device,
        interfaces: List[models.Interface],
        neighbours: Optional[List[models.NetboxDevice]],
        storage: Storage,
) -> models.NetboxDevice:
    return extend_device_base(
        device=device,
        interfaces=interfaces,
        breed=get_breed(
            device.device_type.manufacturer.name,
            device.device_type.model,
        ),
        hw=get_hw(
            device.device_type.manufacturer.name,
            device.device_type.model,
        ),
        neighbours=neighbours,
        storage=storage,
    )


@impl_converter
def extend_interface(
        interface: api_models.Interface,
        ip_addresses: List[models.IpAddress],
) -> models.Interface:
    ...


@impl_converter
def extend_ip_address(
        ip_address: models.IpAddress, prefix: Optional[models.Prefix],
) -> models.IpAddress:
    ...


class NetboxStorageV37(Storage):
    def __init__(self, opts: Optional[NetboxStorageOpts] = None):
        self.netbox = NetboxV37(
            url=opts.url,
            token=opts.token,
        )

    def __enter__(self):
        return self

    def __exit__(self, _, __, ___):
        pass

    def resolve_object_ids_by_query(self, query: NetboxQuery):
        return [
            d.id for d in self._load_devices(query)
        ]

    def resolve_fdnds_by_query(self, query: NetboxQuery):
        return [
            d.name for d in self._load_devices(query)
        ]

    def make_devices(
            self,
            query: Union[NetboxQuery, list],
            preload_neighbors=False,
            use_mesh=None,
            preload_extra_fields=False,
            **kwargs,
    ) -> List[models.NetboxDevice]:
        if isinstance(query, list):
            query = NetboxQuery.new(query)
        device_ids = {
            device.id: extend_device(
                device=device,
                interfaces=[],
                neighbours=[],
                storage=self,
            )
            for device in self._load_devices(query)
        }
        if not device_ids:
            return []

        interfaces = self._load_interfaces(list(device_ids))
        neighbours = {x.id: x for x in self._load_neighbours(interfaces)}
        neighbours_seen = defaultdict(set)

        for interface in interfaces:
            device_ids[interface.device.id].interfaces.append(interface)
            for e in interface.connected_endpoints or []:
                neighbour = neighbours[e.device.id]
                if neighbour.id not in neighbours_seen[interface.device.id]:
                    neighbours_seen[interface.device.id].add(neighbour.id)
                    device_ids[interface.device.id].neighbours.append(neighbour)

        return list(device_ids.values())

    def _load_devices(self, query: NetboxQuery) -> List[api_models.Device]:
        if not query.globs:
            return []
        return [
            device
            for device in self.netbox.dcim_all_devices(
                name__ic=query.globs,
            ).results
            if _match_query(query, device)
            if is_supported(device.device_type.manufacturer.name)
        ]

    def _extend_interfaces(self, interfaces: List[models.Interface]) -> List[models.Interface]:
        extended_ifaces = {
            interface.id: extend_interface(interface, [])
            for interface in interfaces
        }

        ips = self.netbox.ipam_all_ip_addresses(interface_id=list(extended_ifaces))
        ip_to_cidrs: Dict[str, str] = {ip.address: str(ip_interface(ip.address).network) for ip in ips.results}
        prefixes = self.netbox.ipam_all_prefixes(prefix=list(ip_to_cidrs.values()))
        cidr_to_prefix: Dict[str, models.Prefix] = {x.prefix: x for x in prefixes.results}

        for ip in ips.results:
            cidr = ip_to_cidrs[ip.address]
            ip = extend_ip_address(ip, prefix=cidr_to_prefix.get(cidr))
            extended_ifaces[ip.assigned_object_id].ip_addresses.append(ip)
        return list(extended_ifaces.values())

    def _load_interfaces(self, device_ids: List[int]) -> List[
        models.Interface]:
        interfaces = self.netbox.dcim_all_interfaces(device_id=device_ids)
        return self._extend_interfaces(interfaces.results)

    def _load_interfaces_by_id(self, ids: List[int]) -> List[models.Interface]:
        interfaces = self.netbox.dcim_all_interfaces_by_id(id=ids)
        return self._extend_interfaces(interfaces.results)

    def _load_neighbours(self, interfaces: List[models.Interface]) -> List[models.NetboxDevice]:
        endpoints = [e for i in interfaces for e in i.connected_endpoints or []]
        remote_interfaces_ids = [e.id for e in endpoints]
        neighbours_ids = [e.device.id for e in endpoints]
        neighbours_ifaces_dics = defaultdict(list)
        # load only the connected interface to speed things up
        for iface in self._load_interfaces_by_id(remote_interfaces_ids):
            neighbours_ifaces_dics[iface.device.id].append(iface)
        neighbours = []
        for neighbour in self.netbox.dcim_all_devices_by_id(id=neighbours_ids).results:
            extended_neighbour = extend_device(
                device=neighbour,
                storage=self,
                interfaces=neighbours_ifaces_dics[neighbour.id],
                neighbours=None,  # do not load recursively
            )
            neighbours.append(extended_neighbour)
        return neighbours

    def get_device(
            self, obj_id, preload_neighbors=False, use_mesh=None,
            **kwargs,
    ) -> models.NetboxDevice:
        device = self.netbox.dcim_device(obj_id)
        interfaces = self._load_interfaces([device.id])
        neighbours = self._load_neighbours(interfaces)

        res = extend_device(
            device=device,
            storage=self,
            interfaces=interfaces[device.id],
            neighbours=neighbours,
        )
        return res

    def flush_perf(self):
        pass


def _match_query(query: NetboxQuery, device_data: api_models.Device) -> bool:
    for subquery in query.globs:
        if subquery.strip() in device_data.name:
            return True
    return False



# pylint: disable=unused-argument


import abc
import itertools
import re
from collections import namedtuple
from contextlib import contextmanager
from typing import Dict, List, Optional, Type, Any, OrderedDict

from contextlog import get_logger

from annet import text_term_format
from annet.annlib.command import Command, Question, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.annlib.rbparser.deploying import MakeMessageMatcher, Answer
from annet.cli_args import DeployOptions
from annet.connectors import Connector
from annet.output import TextArgs
from annet.rulebook import get_rulebook, deploying
from annet.storage import Device


NCURSES_SIZE_T = 2 ** 15 - 1


_DeployResultBase = namedtuple("_DeployResultBase", ("hostnames", "results", "durations", "original_states"))


class DeployResult(_DeployResultBase):  # noqa: E302
    def add_results(self, results: Dict[str, Optional[Exception]]) -> None:
        for hostname, result in results.items():
            self.hostnames.append(hostname)
            self.results[hostname] = result
            self.durations[hostname] = 0.0
            self.original_states[hostname] = None


class _FetcherConnector(Connector["Fetcher"]):
    name = "Fetcher"
    ep_name = "deploy_fetcher"

    def _get_default(self) -> Type["Fetcher"]:
        return StubFetcher


class _DriverConnector(Connector["DeployDriver"]):
    name = "DeployDriver"
    ep_name = "deploy_driver"

    def _get_default(self) -> Type["DeployDriver"]:
        return StubDeployDriver


fetcher_connector = _FetcherConnector()
driver_connector = _DriverConnector()


class Fetcher(abc.ABC):
    @abc.abstractmethod
    def fetch_packages(self, devices: List[Device],
                       processes: int = 1, max_slots: int = 0):
        pass

    @abc.abstractmethod
    def fetch(self, devices: List[Device],
              files_to_download: Dict[str, List[str]] = None,
              processes: int = 1, max_slots: int = 0):
        pass


class StubFetcher(Fetcher):
    def fetch_packages(self, devices: List[Device],
                       processes: int = 1, max_slots: int = 0):
        raise NotImplementedError()

    def fetch(self, devices: List[Device],
              files_to_download: Dict[str, List[str]] = None,
              processes: int = 1, max_slots: int = 0):
        raise NotImplementedError()


class DeployDriver(abc.ABC):
    @abc.abstractmethod
    async def bulk_deploy(self, deploy_cmds: dict, args: DeployOptions) -> DeployResult:
        pass

    @abc.abstractmethod
    def apply_deploy_rulebook(self, hw, cmd_paths, do_finalize=True, do_commit=True):
        pass

    @abc.abstractmethod
    def build_configuration_cmdlist(self, hw, do_finalize=True, do_commit=True):
        pass

    @abc.abstractmethod
    def build_exit_cmdlist(self, hw):
        pass


class StubDeployDriver(DeployDriver):
    async def bulk_deploy(self, deploy_cmds: dict, args: DeployOptions) -> DeployResult:
        NotImplementedError()

    def apply_deploy_rulebook(self, hw, cmd_paths, do_finalize=True, do_commit=True):
        NotImplementedError()

    def build_configuration_cmdlist(self, hw, do_finalize=True, do_commit=True):
        NotImplementedError()

    def build_exit_cmdlist(self, hw):
        raise NotImplementedError()


# ===
def scrub_config(text, breed):
    return text


def show_bulk_report(hostnames, results, durations, log_dir):
    pass


class AskConfirm:
    CUT_WARN_MSG = "WARNING: the text was cut because of curses limits."

    def __init__(self, text: str, text_type="diff", alternative_text: str = "",
                 alternative_text_type: str = "diff", allow_force_yes: bool = False):
        self.text = [text, text_type]
        self.alternative_text = [alternative_text, alternative_text_type]
        self.color_to_curses: Dict[Optional[str], int] = {}
        self.lines: Dict[int, List[TextArgs]] = {}
        self.rows = None
        self.cols = None
        self.top = 0
        self.left = 0
        self.pad = None
        self.screen = None
        self.found_pos = {}
        self.curses_lines = None
        self.debug_prompt = TextArgs("")
        self.page_position = TextArgs("")
        s_force = "/f" if allow_force_yes else ""
        self.prompt = [
            TextArgs("Execute these commands? [Y%s/q] (/ - search, a - patch/cmds)" % s_force, "blue", offset=0),
            self.page_position,
            self.debug_prompt]

    def _parse_text(self):
        txt = self.text[0]
        txt_split = txt.splitlines()
        # curses pad, который тут используется, имеет ограничение на количество линий
        if (len(txt_split) + 1) >= NCURSES_SIZE_T:  # +1 для того чтобы курсор можно было переместить на пустую строку
            del txt_split[NCURSES_SIZE_T - 3:]
            txt_split.insert(0, self.CUT_WARN_MSG)
            txt_split.append(self.CUT_WARN_MSG)
            txt = "\n".join(txt_split)
        self.rows = len(txt_split)
        self.cols = max(len(line) for line in txt_split)
        res = text_term_format.curses_format(txt, self.text[1])
        self.lines = res

    def _update_search_pos(self, expr):
        self.found_pos = {}
        if not expr:
            return
        try:
            expr = re.compile(expr)
        except Exception:
            return None
        lines = self.text[0].splitlines()
        for (line_no, line) in enumerate(lines):
            for match in re.finditer(expr, line):
                if line_no not in self.found_pos:
                    self.found_pos[line_no] = []
                self.found_pos[line_no].append(TextArgs(match.group(0), "highlight", match.start()))

    def _init_colors(self):
        self.color_to_curses = init_colors()

    def _init_pad(self):
        import curses

        with self._store_xy():
            self.pad = curses.newpad(self.rows + 1, self.cols)
            self.pad.keypad(True)  # accept arrow keys
            self._render_to_pad(self.lines)

    def _render_to_pad(self, lines: dict):
        """
        Рендерим данный на pad
        :param lines: словарь проиндексированный по номерам линий
        :return:
        """
        with self._store_xy():
            for line_no, line_data in sorted(lines.items()):
                line_pos_calc = 0
                for line_part in line_data:
                    if line_part.offset is not None:
                        line_pos = line_part.offset
                    else:
                        line_pos = line_pos_calc
                    if line_part.color:
                        self.pad.addstr(line_no, line_pos, line_part.text, self.color_to_curses[line_part.color])
                    else:
                        self.pad.addstr(line_no, line_pos, line_part.text)
                    line_pos_calc += len(line_part.text)

    def _add_prompt(self):
        for prompt_part in self.prompt:
            if not prompt_part:
                continue
            if prompt_part.offset is None:
                offset = 0
            else:
                offset = prompt_part.offset
            self.screen.addstr(self.curses_lines - 1, offset, prompt_part.text, self.color_to_curses[prompt_part.color])

    def _clear_prompt(self):
        with self._store_xy():
            self.screen.move(self.curses_lines - 1, 0)
            self.screen.clrtoeol()

    def show(self):
        self._add_prompt()
        self.screen.refresh()
        size = self.screen.getmaxyx()
        self.pad.refresh(self.top, self.left, 0, 0, size[0] - 2, size[1] - 2)

    @contextmanager
    def _store_xy(self):
        if self.pad is not None:
            current_y, current_x = self.pad.getyx()
            yield current_y, current_x
            max_y, max_x = self.pad.getmaxyx()
            current_y = min(max_y - 1, current_y)
            current_x = min(max_x - 1, current_x)

            self.pad.move(current_y, current_x)
        else:
            yield

    def search_next(self, prev=False):
        to = None
        current_y, current_x = self.pad.getyx()
        if prev:
            for line_index in sorted(self.found_pos, reverse=True):
                for text_args in self.found_pos[line_index]:
                    if line_index > current_y:
                        continue

                    if line_index < current_y or line_index == current_y and text_args.offset < current_x:
                        to = line_index, text_args.offset
                        break
                if to:
                    break
        else:
            for line_index in sorted([i for i in self.found_pos if i >= current_y]):
                for text_args in self.found_pos[line_index]:
                    if line_index > current_y or line_index == current_y and text_args.offset > current_x:
                        to = line_index, text_args.offset
                        break
                if to:
                    break
        if to:
            return to[0] - current_y, to[1] - current_x
        else:
            return 0, 0

    def _search_prompt(self):
        import curses

        search_prompt = [TextArgs("Search: ", "green_bold", offset=0)]
        current_prompt = self.prompt
        self.prompt = search_prompt
        with self._store_xy():
            self._clear_prompt()
            self.show()
            curses.echo()
            expr = self.screen.getstr().decode()
            curses.noecho()
            self._update_search_pos(expr)
            self._parse_text()
            self._init_pad()
            # срендерем поверх pad слой с подстветкой
            self._render_to_pad(self.found_pos)
            y_offset, x_offset = self.search_next()
        self.prompt = current_prompt
        return y_offset, x_offset

    def _do_commands(self):
        import curses

        while True:
            self._clear_prompt()
            try:
                ch = self.pad.getch()
            except KeyboardInterrupt:
                return "n"
            max_y, max_x = self.screen.getmaxyx()
            _, pad_max_x = self.pad.getmaxyx()
            max_y -= 2  # prompt
            y_offset = 0
            x_offset = 0
            margin = 0
            y_delta = 0
            x_delta = 0

            y, x = self.pad.getyx()
            if ch == ord("q"):
                return "exit"
            elif ch in [ord("y"), ord("Y")]:
                return "y"
            elif ch in [ord("f"), ord("F")]:
                return "force-yes"
            elif ch == ord("a"):
                if self.alternative_text:
                    self.text, self.alternative_text = self.alternative_text, self.text
                self.screen.clear()
                self._parse_text()
                self._init_pad()
            elif ch == ord("d"):
                if self.debug_prompt.text == "":
                    self.debug_prompt.text = "init"
                else:
                    self.debug_prompt.text = ""
            elif ch == ord("n"):
                y_offset, x_offset = self.search_next()
                margin = 10
            elif ch == ord("N"):
                y_offset, x_offset = self.search_next(prev=True)
                margin = 10
            elif ch == ord("/"):
                y_offset, x_offset = self._search_prompt()
                margin = 10
            elif ch == curses.KEY_UP:
                y_offset = -1
            elif ch == curses.KEY_PPAGE:
                y_offset = -10
            elif ch == curses.KEY_HOME:
                y_offset = -len(self.lines)
            elif ch == curses.KEY_DOWN:
                y_offset = 1
            elif ch == curses.KEY_NPAGE:
                y_offset = 10
            elif ch == curses.KEY_END:
                y_offset = len(self.lines)
            elif ch == curses.KEY_LEFT:
                x_offset = -1
            elif ch == curses.KEY_RIGHT:
                x_offset = 1

            if y_offset or x_offset:
                y = max(0, y + y_offset)
                y = min(self.rows, y)
                x = max(0, x + x_offset)
                x = min(self.cols, x)

                y_delta = y - (self.top + max_y - margin)
                if y_delta > 0:
                    self.top += y_delta
                elif (y - margin) < self.top:
                    self.top = y

                self.top = min(self.top, len(self.lines) - max_y)

                x_delta = x - (self.left + max_x)
                if x_delta > 0:
                    self.left += x_delta
                elif x < self.left:
                    self.left = x

                x = min(x, pad_max_x - 1)
                self.pad.move(y, x)

            if self.debug_prompt.text != "":
                debug_line = "y=%s x=%s, x_delta=%s y_delta=%s top=%s, max_y=%s max_x=%s lines=%s" % \
                             (y, x, x_delta, y_delta, self.top, max_y, max_x, len(self.lines))
                self.debug_prompt.text = debug_line
                self.debug_prompt.color = "green_bold"
                self.debug_prompt.offset = max_x - len(debug_line) - 1

            if self.debug_prompt.text == "":
                self.page_position.color = "highlight"
                self.page_position.text = "line %s/%s" % (y, len(self.lines))
                self.page_position.offset = max_x - len(self.page_position.text) - 1

            self.show()

    def loop(self):
        import curses

        res = None
        old_cursor = None
        try:
            self.screen = curses.initscr()
            self.screen.leaveok(True)
            self.curses_lines = curses.LINES  # pylint: disable=maybe-no-member
            curses.start_color()
            curses.noecho()  # no echo key input
            curses.cbreak()  # input with no-enter keyed
            try:
                old_cursor = curses.curs_set(2)
            except Exception:
                pass
            self._init_colors()
            self._parse_text()
            self._init_pad()
            self.pad.move(0, 0)
            self.show()
            res = self._do_commands()
        except Exception as err:
            get_logger().exception("%s", err)
        finally:
            if old_cursor is not None:
                curses.curs_set(old_cursor)
            curses.nocbreak()
            curses.echo()
            curses.endwin()
        return res


def init_colors():
    import curses

    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_CYAN, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
    curses.init_pair(5, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(6, curses.COLOR_BLUE, curses.COLOR_WHITE)
    curses.init_pair(7, curses.COLOR_RED, curses.COLOR_WHITE)
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE)
    curses.init_pair(9, curses.COLOR_CYAN, curses.COLOR_BLUE)
    return {
        "green": curses.color_pair(1),
        "green_bold": curses.color_pair(1) | curses.A_BOLD,
        "cyan": curses.color_pair(2),
        "red": curses.color_pair(3),
        "magenta": curses.color_pair(4),
        "yellow": curses.color_pair(5),
        "blue": curses.color_pair(6),
        "highlight": curses.color_pair(7),
        None: curses.color_pair(8),
        "cyan_blue": curses.color_pair(9),
    }


class RulebookQuestionHandler:
    def __init__(self, dialogs):
        self._dialogs = dialogs

    def __call__(self, dev: Connector, cmd: Command, match_content: bytes):
        content = match_content.strip()
        content = content.decode()
        for matcher, answer in self._dialogs.items():
            if matcher(content):
                return Command(answer.text)

        get_logger().info("no answer in rulebook. dialogs=%s match_content=%s", self._dialogs, match_content)
        return None


def rb_question_to_question(q: MakeMessageMatcher, a: Answer) -> Question:  # TODO: drop MakeMessageMatcher
    if not a.send_nl:
        raise Exception("not supported false send_nl")
    text: str = q._text  # pylint: disable=protected-access
    is_regexp = False
    if text.startswith("/") and text.endswith("/"):
        is_regexp = True
        text = text[1:-1]
    res = Question(question=text, answer=a.text, is_regexp=is_regexp)
    return res


def make_cmd_params(rule: Dict[str, Any]) -> Dict[str, Any]:
    if rule:
        qa_handler = RulebookQuestionHandler(rule["attrs"]["dialogs"])
        qa_list: List[Question] = []
        for matcher, answer in qa_handler._dialogs.items():  # pylint: disable=protected-access
            qa_list.append(rb_question_to_question(matcher, answer))
        return {
            "questions": qa_list,
            "timeout": rule["attrs"]["timeout"],
        }
    return {
        "timeout": 30,
    }


def make_apply_commands(rule, hw, do_commit, do_finalize):
    apply_logic = rule["attrs"]["apply_logic"]
    before, after = apply_logic(hw, do_commit=do_commit, do_finalize=do_finalize)
    return before, after


def fill_cmd_params(rules: OrderedDict, cmd: Command):
    rule = deploying.match_deploy_rule(rules, (cmd.cmd,), {})
    if rule:
        cmd_params = make_cmd_params(rule)
        cmd.questions = cmd_params.get("questions", None)
        cmd.timeout = cmd_params["timeout"]


def apply_deploy_rulebook(hw: HardwareView, cmd_paths, do_finalize=True, do_commit=True):
    rules = get_rulebook(hw)["deploying"]
    cmds_with_apply = []
    for cmd_path, context in cmd_paths.items():
        rule = deploying.match_deploy_rule(rules, cmd_path, context)
        cmd_params = make_cmd_params(rule)
        before, after = make_apply_commands(rule, hw, do_commit, do_finalize)

        cmd = Command(cmd_path[-1], **cmd_params)
        # XXX более чистый способ передавать-мета инфу о команде
        cmd.level = len(cmd_path) - 1
        cmds_with_apply.append((cmd, before, after))

    def _key(item):
        _cmd, before, after = item
        return (tuple(cmd.cmd for cmd in before), tuple(cmd.cmd for cmd in after))

    cmdlist = CommandList()
    for _k, cmd_before_after in itertools.groupby(cmds_with_apply, key=_key):
        cmd_before_after = list(cmd_before_after)
        _, before, after = cmd_before_after[0]
        for c in before:
            c.level = 0
            fill_cmd_params(rules, c)
            cmdlist.add_cmd(c)
        for cmd, _before, _after in cmd_before_after:
            cmdlist.add_cmd(cmd)
        for c in after:
            c.level = 0
            fill_cmd_params(rules, c)
            cmdlist.add_cmd(c)
    return cmdlist



import enum
from typing import List, Tuple

Diff = List[Tuple]


# Операции отмечающие роль команды в дифе
# XXX надо отдельно переделать на enum
class Op:
    ADDED = "added"
    REMOVED = "removed"
    AFFECTED = "affected"
    MOVED = "moved"
    UNCHANGED = "unchanged"


class GeneratorType(enum.Enum):
    PARTIAL = "partial"
    ENTIRE = "entire"
    JSON_FRAGMENT = "json_fragment"

    @staticmethod
    def fromstring(value: str) -> "GeneratorType":
        return GeneratorType(value)

    def tostring(self) -> str:
        return self.value

    def __lt__(self, other: "GeneratorType") -> bool:
        return self.value < other.value

    def __le__(self, other: "GeneratorType") -> bool:
        if self != other:
            return self.value < other.value
        return True
